(** Time-stamp: <modified the 27/03/2008 (at 17:40) by Erwan Jahier> *)


open Lxm
open Errors
open SyntaxTree
open SyntaxTreeCore
open CompiledData


let finish_me msg = print_string ("\n\tXXX LazyCompiler:"^msg^" ->  finish me!\n")

(******************************************************************************)
(** Returns the ident on which the recursion was detected, plus an execution
    stack description. 
*)
exception Recursion_error of (Ident.long as 'id) * (string list as 'stack)
  
exception BadCheckRef_error

let recursion_error (lxm : Lxm.t) (stack : string list) =
  let rec string_of_stack = function
    | [] -> "nostack" 
    | s::[] -> s
    | s::l  -> s^"\n   > "^(string_of_stack l)
  in
    raise ( Compile_error (lxm,
			   "recursion loop detected:\n   > "
			   ^(string_of_stack stack)
   ))


(******************************************************************************)
(* Structure principale *)
type t = {
  src_tab : SyntaxTab.t;
  (* table des defs *)
  types  : (item_key, CompiledData.type_eff     check_flag) Hashtbl.t;
  consts : (item_key, CompiledData.const_eff    check_flag) Hashtbl.t;
  nodes  : (node_key, CompiledData.node_exp_eff check_flag) Hashtbl.t;
  (* table des prov *)
  prov_types  : (item_key, CompiledData.type_eff     check_flag) Hashtbl.t;
  prov_consts : (item_key, CompiledData.const_eff    check_flag) Hashtbl.t;
  prov_nodes  : (node_key, CompiledData.node_exp_eff check_flag) Hashtbl.t
}

(******************************************************************************)
(* exported *)

let (create : SyntaxTab.t -> t) =
fun tbl -> 
  {
    src_tab = tbl;
    types = Hashtbl.create 0;
    consts =  Hashtbl.create 0;
    nodes  = Hashtbl.create 0;
    prov_types = Hashtbl.create 0;
    prov_consts =  Hashtbl.create 0;
    prov_nodes  = Hashtbl.create 0;

(* XXX Remplir ces tables avec les infos relatives aux oprateurs prdfinis !!! *)

(* XXX il manque aussi une table pour les clocks !!! *)

  } 

(******************************************************************************)


(******************************************************************************)

(** Type checking + constant checking/evaluation

   This is performed (lazily) by 10 mutually recursive functions:

   checking types 
   --------------
   (1) [type_check env type_name lxm]: type check the type id [type_name]
   (2) [type_check_do]: untabulated version of [type_check] (do the real work).

   (3) [type_check_interface]: ditto, but for the interface part
   (4) [type_check_interface_do]: untabulated version (do the real work)

   (5) [solve_type_idref] solves constant reference (w.r.t. short/long ident)

   checking constants 
   ------------------
   (6) [const_check env const_name lxm]: eval/check the constant [const_name]
   (7) [const_check_do] : untabulated version (do the real work)

   (8) [const_check_interface]: ditto, but for the interface part
   (9) [const_check_interface_do]: untabulated version (do the real work)

   (10) [solve_const_idref] solves constant reference (w.r.t. short/long ident)

   checking nodes 
   --------------

   (11) [node_check env node_name lxm]: check the node [node_name]
    checking a node means checking its interface and checking it equations/asserts.
    checking an equation means checking that the type and clock of the
    left part is the same as the ones of the rigth part.


   (12) [node_check_do] : untabulated version (do the real work)

   (13) [node_check_interface]: ditto, but for the interface part
   (14) [node_check_interface_do]: untabulated version (do the real work)

   (15) [solve_node_idref] solves constant reference (w.r.t. short/long ident)

    XXX checking clocks 
    -------------------
    Ditto, but todo!


    nb: for x in {type, const, node, clock}, there are several functions 
    that returns [x_eff]:
    - [x_check]
        o tabulates its result
        o takes an x_key and returns an [x_eff]
        o lookups its (syntaxic) definition (x_info) via the symbolTab.t
        o calls [GetEff.X] to translate its sub-terms

    - [GetEff.X]
        o takes a [x_exp] (i.e., an expression) and returns an [x_eff]
        o compute the effective static args (for nodes)
        o calls [solve_x_idref] (via [id_solver]) to translate its sub-terms


    - [solve_x_idref]
        o takes an idref (plus a static_arg_eff list for x=node!)
        o perform name resolution
        o calls [x_check] (loop!)


    nb2: the top-level call is [node_check], on a node that necessarily contains
    no static parameters.  
  

*)

(* Before starting, let's define a few utilitary functions. *)

(** Intermediary results are put into a table. This tabulation handling
    is common to type and constant checking, and is performed by the
    2 following functions.

    Since [x] is meant to stand for [type], [const], or [node], those 2
    functions will lead to the definition of 6 functions:
    [type_check], [const_check], [node_check],
    [type_check_interface], [const_check_interface], [node_check_interface].
*)
let x_check 
    tab find_x x_check_do lookup_x_eff pack_of_x_key name_of_x_key this x_key lxm =
  try lookup_x_eff tab x_key lxm 
  with Not_found ->
    Hashtbl.add tab x_key Checking;
    let (x_pack,xn) = (pack_of_x_key x_key, name_of_x_key x_key) in
    let x_pack_symbols = SyntaxTab.pack_body_env this.src_tab x_pack in
    let x_def = match find_x x_pack_symbols xn lxm with
      | SymbolTab.Here x_def -> x_def
      | SymbolTab.NotHere _ -> assert false (* should not occur *)
    in
    let res = x_check_do this x_key lxm x_pack_symbols x_pack x_def in
      Hashtbl.replace tab x_key (Checked res);
      res

let x_check_interface 
    tab find_x x_check x_check_interface_do lookup_x_eff 
    pack_of_x_key name_of_x_key this x_key lxm =
  try lookup_x_eff tab x_key lxm
  with Not_found ->
    Hashtbl.add tab x_key Checking;
    let (xp,xn) = (pack_of_x_key x_key, name_of_x_key x_key) in
    let xp_prov_symbols_opt = SyntaxTab.pack_prov_env this.src_tab xp in
    let res = (* [xp] migth have no provided symbol table *)
      match xp_prov_symbols_opt with
	| None -> x_check this x_key lxm
	    (* if [xp] have no provided symbol table, the whole
	       package is exported. *)
	| Some xp_prov_symbols ->
	    let x_def = match find_x xp_prov_symbols xn lxm with
	      | SymbolTab.Here x -> x
	      | SymbolTab.NotHere _ -> assert false (* should not occur *)
	    in
	      x_check_interface_do this x_key lxm xp_prov_symbols xp x_def
    in
      Hashtbl.replace tab x_key (Checked res);
      res

(* Returns the tabulated [type] or [const], if it has already been computed;
   otherwise, raise [Not_found] otherwise. *)
let lookup_x_eff x_label id_of_x_key x_tab x_key lxm  =
    match Hashtbl.find x_tab x_key with
      | Checked res -> res
      | Checking -> 
	  raise (Recursion_error (id_of_x_key x_key, [x_label^(Lxm.details lxm)]))
      | Incorrect -> raise (BadCheckRef_error)

let (lookup_type_eff: (item_key, CompiledData.type_eff check_flag) Hashtbl.t -> 
      Ident.long -> Lxm.t -> CompiledData.type_eff) = 
  lookup_x_eff "type ref "  (fun k -> k)

let (lookup_const_eff:(item_key, CompiledData.const_eff check_flag) Hashtbl.t -> 
      Ident.long -> Lxm.t -> CompiledData.const_eff) = 
  lookup_x_eff "const ref " (fun k -> k)

let (lookup_node_exp_eff:
       (CompiledData.node_key, CompiledData.node_exp_eff check_flag) Hashtbl.t -> 
      CompiledData.node_key -> Lxm.t -> CompiledData.node_exp_eff) = 
  lookup_x_eff "node ref "  (fun k -> fst k)


(** From an idref, builds a [x_key] and calls [x_check] *)
let solve_x_idref
    x_check_interface x_check find_x x_label to_x_key this symbols currpack idr 
    sargs lxm =
  let s = Ident.name_of_idref idr in
    match Ident.pack_of_idref idr with
      | Some p -> x_check_interface this (to_x_key p s) lxm
      | None ->
	  (* no pack name: we look it up in the symbols table *)
	  try
	    match (find_x symbols s lxm) with
	      | SymbolTab.Here _      -> x_check this (to_x_key currpack s) lxm
	      | SymbolTab.NotHere fid -> 
		  x_check_interface this 
		    (to_x_key (Ident.pack_of_long fid) (Ident.of_long fid)) 
		    lxm
	  with Not_found -> 
	    (raise (Compile_error(lxm,"unbounded " ^ x_label ^ " ident")))


(* And now we can start the big mutually recursive definition... *)

(** Tabulated version of [type_check_do]. *)
let rec (type_check : t -> Ident.long -> Lxm.t -> CompiledData.type_eff) =
  fun this -> 
    x_check this.types SymbolTab.find_type type_check_do lookup_type_eff 
      Ident.pack_of_long Ident.of_long this

(** Tabulated version of [const_check_do]. *)
and (const_check : t -> Ident.long -> Lxm.t -> CompiledData.const_eff) =
  fun this -> 
  x_check this.consts SymbolTab.find_const const_check_do lookup_const_eff 
    Ident.pack_of_long Ident.of_long this

(** Tabulated version of [type_check_interface_do]. *)
and (type_check_interface: t -> Ident.long -> Lxm.t -> CompiledData.type_eff) =
  fun this -> 
  x_check_interface 
    this.prov_types SymbolTab.find_type type_check type_check_interface_do 
    lookup_type_eff Ident.pack_of_long Ident.of_long this

(** Tabulated version of [const_check_interface_do]. *)
and (const_check_interface: t -> Ident.long -> Lxm.t -> CompiledData.const_eff) =
  fun this -> 
    x_check_interface 
      this.prov_consts SymbolTab.find_const const_check const_check_interface_do
      lookup_const_eff Ident.pack_of_long Ident.of_long this

(** solving type and constant references *)
and (solve_type_idref : t -> SymbolTab.t -> Ident.pack_name -> Ident.idref -> Lxm.t
      -> CompiledData.type_eff) =
  fun this symbols currpack idr lxm -> 
    solve_x_idref 
      type_check_interface type_check SymbolTab.find_type "type"
      (fun p id -> Ident.make_long p id)
      this symbols currpack idr [] lxm

and (solve_const_idref : t -> SymbolTab.t -> Ident.pack_name -> Ident.idref -> Lxm.t
      -> CompiledData.const_eff) =
  fun this symbols currpack idr lxm ->
    solve_x_idref
      const_check_interface const_check SymbolTab.find_const "const"
      (fun p id -> Ident.make_long p id)
      this symbols currpack idr [] lxm


(* now the real work! *)
and (type_check_interface_do: t -> Ident.long -> Lxm.t -> SymbolTab.t ->
      Ident.pack_name -> SyntaxTreeCore.type_info srcflagged -> 
      CompiledData.type_eff) =
  fun this type_name lxm prov_symbols pack_name type_def ->
    (* We type check the interface and the body. 
       For non-abstract types, we also check that both effective types are
       the same.  *)
    let body_type_eff = type_check this type_name lxm in
    let prov_type_eff =
      type_check_do this type_name lxm prov_symbols pack_name type_def
    in
      match prov_type_eff with
	| External_type_eff _ -> prov_type_eff (* no check is necessary *)
	| Bool_type_eff
	| Int_type_eff
	| Real_type_eff
	| Enum_type_eff (_,_)
	| Array_type_eff (_,_)
	| Struct_type_eff (_,_)
	  -> (* Not an abstract type: therefore, we check that [prov_type_eff]
		is the same as [body_type_eff].
	     *)
	    let body_symbols = SyntaxTab.pack_body_env this.src_tab pack_name in
	    let prov_type_eff_body = (* rm? *)
	      type_check_do this type_name lxm body_symbols pack_name type_def
	    in
	      if prov_type_eff_body = body_type_eff then
		prov_type_eff
	      else
		raise(Compile_error (
			type_def.src,
			("provided type \n\t" ^ 
			   (string_of_type_eff prov_type_eff_body) ^
			   "\n is not compatible with its implementation \n\t" ^ 
			   (string_of_type_eff body_type_eff))))


and (const_check_interface_do: t -> Ident.long -> Lxm.t -> SymbolTab.t -> 
      Ident.pack_name -> SyntaxTreeCore.const_info srcflagged -> 
      CompiledData.const_eff) =
  fun this cn lxm prov_symbols p const_def -> 
    let prov_const_eff = const_check_do this cn lxm prov_symbols p const_def in
    let body_const_eff = const_check this cn lxm in
      match prov_const_eff with
	| Enum_const_eff (id, teff_prov)
	| Extern_const_eff (id, teff_prov) ->
	    let teff_body = type_of_const_eff body_const_eff in
	      if (id = cn) && (teff_prov = teff_body) then body_const_eff 
	      else 
		raise(Compile_error (
			const_def.src,
			("provided constant type \n\t" ^ 
			   (string_of_type_eff teff_prov)  ^ 
			   "   is not compatible with its implementation \n\t" ^ 
			   (string_of_type_eff teff_body) ^ "")
		      ))
	| Bool_const_eff _
	| Int_const_eff _ 
	| Real_const_eff _ 
	| Struct_const_eff (_,_)
	| Array_const_eff (_,_)
	  ->
	    if prov_const_eff = body_const_eff then
	      body_const_eff
	    else
	      raise(Compile_error (
		      const_def.src, 
		      " provided constant does not match with its definition."))


and (type_check_do: t -> Ident.long -> Lxm.t -> SymbolTab.t -> Ident.pack_name ->
      SyntaxTreeCore.type_info srcflagged -> CompiledData.type_eff) =
  fun this type_name lxm symbols pack_name type_def -> 
    try (
      (* Solveur d'idref pour les appels  eval_type/eval_const *)
      let id_solver = {
	id2var   = (fun idref lxm -> assert false (* should not occur *)); 
	id2const = solve_const_idref this symbols pack_name;
	id2type  = solve_type_idref this symbols pack_name;
	id2node  = solve_node_idref this symbols pack_name;
	symbols  = symbols;
      }
      in 
	match type_def.it with
	  | ArrayType _ -> finish_me " array handling "; assert false
	  | ExternalType s        -> External_type_eff (Ident.make_long pack_name s)
	  | AliasedType (s, texp) -> GetEff.typ id_solver texp
	  | EnumType (s, clst) -> (
	      let n = Ident.make_long pack_name s in
	      let add_pack_name x = Ident.make_long pack_name x.it in
		Enum_type_eff (n, List.map add_pack_name clst)
	    )
	  | StructType sti -> (
	      let make_field (fname : Ident.t) =
		let field_def = Hashtbl.find sti.st_ftable fname in
		let teff = GetEff.typ id_solver field_def.it.fd_type in
		  match field_def.it.fd_value with
		    | None -> (fname, (teff, None))
		    | Some vexp -> (
			let veff = EvalConst.f id_solver vexp in
			  match veff with
			    | [v] -> (
				let tv = type_of_const_eff v in
				  if (tv = teff) then (fname, (teff, Some v)) else 
				    raise 
				      (Compile_error(field_def.src, Printf.sprintf
		              " this field is declared as '%s' but evaluated as '%s'"
					   (string_of_type_eff teff)
					   (string_of_type_eff tv)))
			      )
			    | [] -> assert false (* should not occur *)
			    | _::_ -> 
				raise (Compile_error(field_def.src,
					  "bad field value: tuple not allowed"))
		      )
	      in
	      let n = Ident.make_long pack_name sti.st_name in
	      let eff_fields = List.map make_field sti.st_flist in
		Struct_type_eff (n, eff_fields)
	    )
    ) 
    with
	(* capte et complete/stoppe les recursions *)
	Recursion_error (root, stack) ->
	  if (root = type_name) then recursion_error type_def.src stack else
	    raise ( Recursion_error (root, ("type ref "^(Lxm.details lxm))::stack))
	      
	      
and (const_check_do : t -> Ident.long -> Lxm.t -> SymbolTab.t -> Ident.pack_name ->
      SyntaxTreeCore.const_info srcflagged -> CompiledData.const_eff) =
  fun this cn lxm symbols currpack const_def ->
    (* [cn] and [lxm] are used for recursion errors. 
       [symbols] is the current symbol table.
    *)
    try (
      (* Solveur d'idref pour les  les appels  eval_type/eval_const *)
      let id_solver = {
	id2var   = (fun idref lxm -> assert false (* should not occur *)); 
	id2const = solve_const_idref this symbols currpack;
	id2type  = solve_type_idref  this symbols currpack;
	id2node  = solve_node_idref  this symbols currpack;
	symbols  = symbols;
      }
      in
	match const_def.it with
	  | ExternalConst (id, texp) ->
	      Extern_const_eff ((Ident.make_long currpack id), 
				GetEff.typ id_solver texp)
	  | EnumConst (id, texp) ->
	      Enum_const_eff ((Ident.make_long currpack id), GetEff.typ id_solver texp)

	  | DefinedConst (id, texp_opt, vexp ) -> (
	      match (EvalConst.f id_solver vexp) with
		| [ceff] -> (
		    match texp_opt with
		      | None -> ceff
		      | Some texp -> (
			  let tdecl = GetEff.typ id_solver texp in
			  let teff =  type_of_const_eff ceff in
			    if (tdecl = teff ) then ceff else 
			      raise 
				(Compile_error (const_def.src, Printf.sprintf
		             " this constant is declared as '%s' but evaluated as '%s'"
				      (string_of_type_eff tdecl)
				      (string_of_type_eff teff)
				   )))
		  )
		| [] -> assert false (* should not occur *)
		| _::_ -> raise (Compile_error(const_def.src, 
				    "bad constant value: tuple not allowed"))
	    )
    ) with Recursion_error (root, stack) -> (
      (* capte et complete/stoppe les recursions *)
      if (root = cn) then recursion_error const_def.src stack else 
	(* on complete la stack *)
	raise (Recursion_error (root, ("const ref "^(Lxm.details lxm))::stack))
    )


(******************************************************************************)
and (node_check_interface_do: t -> CompiledData.node_key -> Lxm.t -> 
	  SymbolTab.t -> Ident.pack_name -> SyntaxTreeCore.node_info srcflagged -> 
	  CompiledData.node_exp_eff) =
  fun this nk lxm symbols pack_name node_def ->
    let body_node_exp_eff = node_check this nk lxm in
    let prov_node_exp_eff = node_check_do this nk lxm symbols pack_name node_def in
      if 
	prov_node_exp_eff.node_key_eff      = body_node_exp_eff.node_key_eff      &&
	prov_node_exp_eff.inlist_eff        = body_node_exp_eff.inlist_eff        &&
	prov_node_exp_eff.outlist_eff       = body_node_exp_eff.outlist_eff       &&
	prov_node_exp_eff.clock_inlist_eff  = body_node_exp_eff.clock_inlist_eff  &&
	prov_node_exp_eff.clock_outlist_eff = body_node_exp_eff.clock_outlist_eff &&
	prov_node_exp_eff.has_mem_eff       = body_node_exp_eff.has_mem_eff       &&
	prov_node_exp_eff.is_safe_eff       = body_node_exp_eff.is_safe_eff       &&
	match prov_node_exp_eff.def_eff, body_node_exp_eff.def_eff with
	  | (AbstractEff,_) -> true
	  | (_,_) -> prov_node_exp_eff.def_eff = body_node_exp_eff.def_eff
      then
	prov_node_exp_eff
      else
	raise(
	  Compile_error (
	    node_def.src,
	    ("provided node \n\t" ^ (string_of_node_exp_eff prov_node_exp_eff) ^
	       "\n is not compatible with its implementation \n\t" ^
	       (string_of_node_exp_eff body_node_exp_eff))))


and (node_check_do: t -> CompiledData.node_key -> Lxm.t -> SymbolTab.t -> 
      Ident.pack_name -> SyntaxTreeCore.node_info srcflagged -> 
      CompiledData.node_exp_eff) =
  fun this nk lxm symbols pack_name node_def ->
    (*
      - verifier les params statiques ?
      - creer une sorte d'environnement local de compilation, qui
      permette de gerer les espaces de noms propres aux noeuds
      (variables, constantes, flots, types) 
      cf CompiledData.local_env pack_env, et world_env
      
    *)
    let local_env = make_local_env nk in
    let node_id_solver = {
      (* a node [node_id_solver] is a [id_solver] where we begin to
	 look into the local environement before looking at the
	 global one.  *)

      id2var = (* var can only be local to the node *)
	(fun id lxm -> 
	   try lookup_var local_env (Ident.string_of_idref id) lxm
	   with Not_found ->
	     print_string ((Lxm.details lxm) ^ 
			     "\n cannot find " ^ (Ident.string_of_idref id) ^ 
			     " into the node variable table ");
	     Hashtbl.iter
	       (fun id vi_eff ->
		  print_string(
		    "\n\t" ^ (Ident.to_string id) ^ " -> " ^
		      (string_of_var_info_eff vi_eff))
	       )
	       local_env.lenv_vars;
	     flush stdout;
	     assert false
	);
      id2const =
	(fun id lxm ->
	   try lookup_const local_env id lxm
	   with Not_found -> solve_const_idref this symbols pack_name id lxm);
      id2type  =
	(fun id lxm ->
	   try lookup_type local_env id lxm
	   with Not_found -> solve_type_idref  this symbols pack_name id lxm);
      id2node  =
	(fun id sargs lxm ->
	   try lookup_node local_env  id sargs lxm
	   with Not_found -> solve_node_idref  this symbols pack_name id sargs lxm);
      symbols  = symbols;
    }
    in
    let make_node_eff node_def_eff =
      (* building not aliased nodes *)
      match node_def.it.vars with
	| None -> assert false (* a node with a body should have a profile *)
	| Some vars ->
	    let type_args id =
	      let vi = Hashtbl.find vars.vartable id in
	      let t_eff = GetEff.typ node_id_solver vi.it.var_type in
	      let vi_eff = {
		   var_name_eff   = vi.it.var_name;
		   var_nature_eff = vi.it.var_nature;
		   var_type_eff   = t_eff;
	      }
	      in
		Hashtbl.add local_env.lenv_types id t_eff;
		Hashtbl.add local_env.lenv_vars id vi_eff;
		t_eff
	    in
	    let inlist =  List.map type_args vars.inlist
	    and outlist = List.map type_args vars.outlist
	    and loclist = 
	      match vars.loclist with
		| None -> None
		| Some loclist -> Some (List.map type_args loclist)
	    in
	      {
		node_key_eff = nk;
		inlist_eff   = inlist;
		outlist_eff  = outlist;
		loclist_eff  = loclist;
		clock_inlist_eff  = [];(* XXX finish me! *)
		clock_outlist_eff = [];(* XXX finish me! *)
		def_eff = node_def_eff ();
		has_mem_eff  = node_def.it.has_mem;
		is_safe_eff  = node_def.it.is_safe;
	      }
    in
    (* let's go *)
      match node_def.it.def with
	| Abstract -> make_node_eff (fun () -> AbstractEff)
	| Extern   -> make_node_eff (fun () -> ExternEff)
	| Body nb ->
	    make_node_eff ( 
	      (fun () -> (* trick to force to delay this evaluation 
			    after the local_env.lenv_vars has been
			    filled
			 *)
		 BodyEff {
		   asserts_eff =
		     List.map (GetEff.val_exp node_id_solver) nb.asserts;
		   eqs_eff = List.map (GetEff.eq node_id_solver) nb.eqs; 
		 }
	      )
	    )
	| Alias node_alias -> (
	    (* just check that the declared profile (if any) matches with the alias *)
	    let res = GetEff.node node_id_solver node_alias in
	      match node_def.it.vars with
		| None -> res
		| Some vars ->
		    let vi_il, vi_ol = 
		      List.map (fun id -> Hashtbl.find vars.vartable id) vars.inlist,
		      List.map (fun id -> Hashtbl.find vars.vartable id) vars.outlist
		    in
		    let aux vi = GetEff.typ node_id_solver vi.it.var_type in
		    let (il,ol) = CompiledData.profile_of_node_exp_eff res in
		      if 
			List.map aux vi_il <> il || List.map aux vi_ol <> ol 
		      then
			raise (Compile_error (
				 lxm, "type mismatch in node alias definition"))
			  (* that error msg could be more precise *)
		      else
			res
	  )


(** builds a [node_key] and calls [node_check] *)
and (solve_node_idref : t -> SymbolTab.t -> Ident.pack_name -> Ident.idref ->
      static_arg_eff list -> Lxm.t -> CompiledData.node_exp_eff) =
  fun this symbols currpack idr sargs lxm ->
    solve_x_idref
      node_check_interface node_check SymbolTab.find_node "node"
      (fun p id ->
	 (* builds a [node_key] from a [pack_name] and a [node] id, 
	 and a static_arg_eff list *)
	 let long = Ident.make_long p id in
	 let node_key = long, sargs in
	   node_key
      )
      this symbols currpack idr sargs lxm

and (node_check: t -> CompiledData.node_key -> Lxm.t -> CompiledData.node_exp_eff) =
  fun this nk ->
    x_check this.nodes SymbolTab.find_node node_check_do lookup_node_exp_eff
      (fun nk -> Ident.pack_of_long (fst nk))
      (fun nk -> Ident.of_long (fst nk))
      this nk

and (node_check_interface:
       t -> CompiledData.node_key -> Lxm.t -> CompiledData.node_exp_eff) =
  fun this nk ->
    x_check_interface this.prov_nodes SymbolTab.find_node node_check
      node_check_interface_do lookup_node_exp_eff
      (fun nk -> Ident.pack_of_long (fst nk))
      (fun nk -> Ident.of_long (fst nk)) this nk
      

    
(*-------------------------------------------------------------------------
 Test/debug
---------------------------------------------------------------------------*)


let test_item this label x_check_interface string_of_x_key string_of_x_eff
    to_key id item_def =
  match item_def with
    | SymbolTab.Here _item_def ->
	let x_eff = x_check_interface this (to_key id) Lxm.dummy in
	  Printf.printf "\t\t%s %s = %s\n"
	    label (string_of_x_key (to_key id)) (string_of_x_eff x_eff)

    | SymbolTab.NotHere item_def ->
	Printf.printf "\t\t%s %s = %s (imported)\n"
	  label (string_of_x_key (to_key id)) (Ident.string_of_long item_def)
   

let test_types pack_name this =
  test_item this "type" type_check_interface Ident.string_of_long 
    string_of_type_eff (fun id -> Ident.make_long pack_name id)

let test_constants pack_name this = 
  test_item this "const" const_check_interface  Ident.string_of_long
    string_of_const_eff  (fun id -> Ident.make_long pack_name id)


let (get_static_params : (node_info Lxm.srcflagged) SymbolTab.hereflagged -> 
      static_param srcflagged list option) =
  fun node_info_flagged -> 
    match node_info_flagged with
      | SymbolTab.Here nif -> 
	  (match nif.it.static_params with 
	     | None -> None
	     | Some [] -> None
	     | Some sp -> Some sp
	  )
      | SymbolTab.NotHere id -> 
	  None (* do not test imported node. there will be tested anyway *)

let test_nodes pack_name this id ni_f =
  match get_static_params ni_f with
    | Some sp -> 
	(* too difficult to test such node easily, so we skip it *)
	Verbose.print_string (" ### skipping " ^ (Ident.to_string id) ^ "\n");
	flush stdout
    | None ->
	test_item this "node" node_check_interface CompiledData.string_of_node_key
	  string_of_node_exp_eff (fun id -> (Ident.make_long pack_name id, [])) id ni_f


let test (this: t) = (
  (* src_tab : SyntaxTab.t; *)
  let testpack pack_name = (
    Verbose.printf " * package %s\n" (Ident.pack_name_to_string pack_name);
    let prov_symbols =
      match SyntaxTab.pack_prov_env this.src_tab pack_name with
	| Some tab -> tab
	| None -> SyntaxTab.pack_body_env this.src_tab pack_name
    in
      Verbose.print_string "\tExported types:\n";
      SymbolTab.iter_types prov_symbols (test_types pack_name this);
      flush stdout;
      Verbose.print_string "\tExported constants:\n";
      SymbolTab.iter_consts prov_symbols (test_constants pack_name this);
      flush stdout;
      Verbose.print_string "\tExported nodes:\n";
      SymbolTab.iter_nodes prov_symbols (test_nodes pack_name this);
      flush stdout
  )
  in
  let plist = SyntaxTab.pack_list this.src_tab in
    Verbose.print_string "*** Dump the exported items of the packages.\n";
    try
      List.iter testpack plist
    with
	Recursion_error (root, stack) -> 
	  recursion_error Lxm.dummy stack
)
